from thrift.transport import TSocket
from thrift.transport import TTransport
from thrift.protocol import TCompactProtocol
from thrift.server import TServer


class ThriftFactory(object):
    """Abstract factory to create Thrift objects for client and server."""

    def __init__(self, transportFactory, protocolFactory):
        self.transportFactory = transportFactory
        self.protocolFactory = protocolFactory

    def createSocket(self, host, port):
        socket = TSocket.TSocket(host, port)
        return socket

    def createTransport(self, socket):
        transport = self.transportFactory.getTransport(socket)
        return transport

    def createProtocol(self, transport):
        protocol = self.protocolFactory.getProtocol(transport)
        return protocol

    def createServer(self, processor, host, port):
        socket = TSocket.TServerSocket(host, port)

        server = TServer.TThreadedServer(processor, socket,
                                         self.transportFactory,
                                         self.protocolFactory)
        return server


factory = ThriftFactory(TTransport.TFramedTransportFactory(),
                        TCompactProtocol.TCompactProtocolAcceleratedFactory())


def is_accelerated():
    try:
        transport = TTransport.TMemoryBuffer()
        TCompactProtocol.TCompactProtocolAccelerated(transport, fallback=False)
        return True
    except:
        return False



from concrete.util.concrete_uuid import (
    generate_UUID, bin_to_hex, hex_to_bin, AnalyticUUIDGeneratorFactory,
    generate_uuid_unif, generate_hex_unif, split_uuid, join_uuid,
)

from concrete import AnnotationMetadata, Communication
from concrete.validate import validate_communication

import re
import time
import unittest

UUID_RE = re.compile(
    r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$')


class TestUUID(unittest.TestCase):

    def test_next(self):
        comm = Communication()
        comm.uuid = generate_UUID()

    def test_minimal_communication_with_uuid(self):
        comm = Communication()
        comm.id = "myID"
        comm.metadata = AnnotationMetadata(
            tool="TEST", timestamp=int(time.time()))
        comm.type = "Test Communication"
        comm.uuid = generate_UUID()
        self.assertTrue(validate_communication(comm))


class TestHexToBin(unittest.TestCase):

    def test_hex_to_bin(self):
        self.assertEqual(hex_to_bin('0'), 0)
        self.assertEqual(hex_to_bin('1'), 1)
        self.assertEqual(hex_to_bin('2'), 2)
        self.assertEqual(hex_to_bin('3'), 3)
        self.assertEqual(hex_to_bin('4'), 4)
        self.assertEqual(hex_to_bin('5'), 5)
        self.assertEqual(hex_to_bin('6'), 6)
        self.assertEqual(hex_to_bin('7'), 7)
        self.assertEqual(hex_to_bin('8'), 8)
        self.assertEqual(hex_to_bin('9'), 9)
        self.assertEqual(hex_to_bin('a'), 10)
        self.assertEqual(hex_to_bin('A'), 10)
        self.assertEqual(hex_to_bin('b'), 11)
        self.assertEqual(hex_to_bin('B'), 11)
        self.assertEqual(hex_to_bin('c'), 12)
        self.assertEqual(hex_to_bin('C'), 12)
        self.assertEqual(hex_to_bin('d'), 13)
        self.assertEqual(hex_to_bin('D'), 13)
        self.assertEqual(hex_to_bin('e'), 14)
        self.assertEqual(hex_to_bin('E'), 14)
        self.assertEqual(hex_to_bin('f'), 15)
        self.assertEqual(hex_to_bin('F'), 15)
        self.assertEqual(hex_to_bin('caf'), 15 + 10 * 16 + 12 * 256)
        self.assertEqual(hex_to_bin('cafe3'), 3 + 14 * 16 +
                         15 * 256 + 10 * 4096 + 12 * 65536)


class TestBinToHex(unittest.TestCase):

    def test_bin_to_hex(self):
        self.assertEqual(bin_to_hex(0), '0')
        self.assertEqual(bin_to_hex(1), '1')
        self.assertEqual(bin_to_hex(2), '2')
        self.assertEqual(bin_to_hex(3), '3')
        self.assertEqual(bin_to_hex(4), '4')
        self.assertEqual(bin_to_hex(5), '5')
        self.assertEqual(bin_to_hex(6), '6')
        self.assertEqual(bin_to_hex(7), '7')
        self.assertEqual(bin_to_hex(8), '8')
        self.assertEqual(bin_to_hex(9), '9')
        self.assertEqual(bin_to_hex(10).lower(), 'a')
        self.assertEqual(bin_to_hex(11).lower(), 'b')
        self.assertEqual(bin_to_hex(12).lower(), 'c')
        self.assertEqual(bin_to_hex(13).lower(), 'd')
        self.assertEqual(bin_to_hex(14).lower(), 'e')
        self.assertEqual(bin_to_hex(15).lower(), 'f')
        self.assertEqual(bin_to_hex(15 + 10 * 16 + 12 * 256).lower(), 'caf')
        self.assertEqual(bin_to_hex(3 + 14 * 16 + 15 * 256 +
                                    10 * 4096 + 12 * 65536).lower(), 'cafe3')

    def test_bin_to_hex_with_len(self):
        self.assertEqual(bin_to_hex(0, 4), '0000')
        self.assertEqual(bin_to_hex(1, 4), '0001')
        self.assertEqual(bin_to_hex(2, 4), '0002')
        self.assertEqual(bin_to_hex(3, 4), '0003')
        self.assertEqual(bin_to_hex(4, 4), '0004')
        self.assertEqual(bin_to_hex(5, 4), '0005')
        self.assertEqual(bin_to_hex(6, 4), '0006')
        self.assertEqual(bin_to_hex(7, 4), '0007')
        self.assertEqual(bin_to_hex(8, 4), '0008')
        self.assertEqual(bin_to_hex(9, 4), '0009')
        self.assertEqual(bin_to_hex(10, 4).lower(), '000a')
        self.assertEqual(bin_to_hex(11, 4).lower(), '000b')
        self.assertEqual(bin_to_hex(12, 4).lower(), '000c')
        self.assertEqual(bin_to_hex(13, 4).lower(), '000d')
        self.assertEqual(bin_to_hex(14, 4).lower(), '000e')
        self.assertEqual(bin_to_hex(15, 4).lower(), '000f')
        self.assertEqual(bin_to_hex(
            15 + 10 * 16 + 12 * 256, 4).lower(), '0caf')
        with self.assertRaises(ValueError):
            bin_to_hex(3 + 14 * 16 + 15 * 256 + 10 * 4096 + 12 * 65536, 4)


class TestSplitUUID(unittest.TestCase):

    def test_split_uuid_valid(self):
        self.assertEqual(
            split_uuid('7575a428-aaf7-4c2e-929e-1e2a0ab59e16'),
            ('7575a428aaf7', '4c2e929e', '1e2a0ab59e16')
        )

    def test_split_uuid_too_many_pieces(self):
        with self.assertRaises(ValueError):
            split_uuid('aaf7-7575a428-aaf7-4c2e-929e-1e2a0ab59e16')

    def test_split_uuid_too_few_pieces(self):
        with self.assertRaises(ValueError):
            split_uuid('7575a428-4c2e-929e-1e2a0ab59e16')

    def test_split_uuid_first_piece_too_long(self):
        with self.assertRaises(ValueError):
            split_uuid('7575a428a-aaf7-4c2e-929e-1e2a0ab59e16')

    def test_split_uuid_second_piece_too_long(self):
        with self.assertRaises(ValueError):
            split_uuid('7575a428-aaf7a-4c2e-929e-1e2a0ab59e16')

    def test_split_uuid_third_piece_too_long(self):
        with self.assertRaises(ValueError):
            split_uuid('7575a428-aaf7-4c2ea-929e-1e2a0ab59e16')

    def test_split_uuid_fourth_piece_too_long(self):
        with self.assertRaises(ValueError):
            split_uuid('7575a428-aaf7-4c2e-929ea-1e2a0ab59e16')

    def test_split_uuid_fifth_piece_too_long(self):
        with self.assertRaises(ValueError):
            split_uuid('7575a428-aaf7-4c2e-929e-1e2a0ab59e16a')

    def test_split_uuid_first_piece_too_short(self):
        with self.assertRaises(ValueError):
            split_uuid('7575a42-aaf7-4c2e-929e-1e2a0ab59e16')

    def test_split_uuid_second_piece_too_short(self):
        with self.assertRaises(ValueError):
            split_uuid('7575a428-aaf-4c2e-929e-1e2a0ab59e16')

    def test_split_uuid_third_piece_too_short(self):
        with self.assertRaises(ValueError):
            split_uuid('7575a428-aaf7-4c2-929e-1e2a0ab59e16')

    def test_split_uuid_fourth_piece_too_short(self):
        with self.assertRaises(ValueError):
            split_uuid('7575a428-aaf7-4c2e-929-1e2a0ab59e16')

    def test_split_uuid_fifth_piece_too_short(self):
        with self.assertRaises(ValueError):
            split_uuid('7575a428-aaf7-4c2e-929e-1e2a0ab59e1')


class TestJoinUUID(unittest.TestCase):

    def test_join_uuid_valid(self):
        self.assertEquals(
            join_uuid('7575a428aaf7', '4c2e929e', '1e2a0ab59e16'),
            '7575a428-aaf7-4c2e-929e-1e2a0ab59e16'
        )

    def test_join_uuid_first_piece_too_long(self):
        with self.assertRaises(ValueError):
            join_uuid('7575a428aaf7a', '4c2e929e', '1e2a0ab59e16')

    def test_join_uuid_second_piece_too_long(self):
        with self.assertRaises(ValueError):
            join_uuid('7575a428aaf7', '4c2e929ea', '1e2a0ab59e16')

    def test_join_uuid_third_piece_too_long(self):
        with self.assertRaises(ValueError):
            join_uuid('7575a428aaf7', '4c2e929e', '1e2a0ab59e16a')

    def test_join_uuid_first_piece_too_short(self):
        with self.assertRaises(ValueError):
            join_uuid('7575a428aaf', '4c2e929e', '1e2a0ab59e16')

    def test_join_uuid_second_piece_too_short(self):
        with self.assertRaises(ValueError):
            join_uuid('7575a428aaf7', '4c2e929', '1e2a0ab59e16')

    def test_join_uuid_third_piece_too_short(self):
        with self.assertRaises(ValueError):
            join_uuid('7575a428aaf7', '4c2e929e', '1e2a0ab59e1')


class TestGenerateHex(unittest.TestCase):

    def test_generate_hex_unif_range(self):
        n = 1000  # (1 - 1/16)^n = 9e-29
        r = set()
        for i in xrange(n):
            r.add(generate_hex_unif(1))
        self.assertEquals(
            sorted(map(lambda x: str(x).lower(), r)),
            [c for c in '0123456789abcdef']
        )

    def test_generate_hex_unif_len(self):
        h = generate_hex_unif(21)  # (1/16)^(n-1) = 8e-25
        self.assertEquals(len(h), 21)
        self.assertTrue(len(set(c for c in h)) > 1)

    def test_generate_hex_unif_spread(self):
        n = 1000
        m = 32
        # union bound: (1/16)^m * n^2 = 3e-33
        s = set([generate_hex_unif(m) for i in xrange(n)])
        self.assertEquals(len(s), n)


class TestGenerateUUID(unittest.TestCase):

    def test_generate_uuid_unif_format(self):
        u = generate_uuid_unif()
        assert UUID_RE.match(u) is not None

    def test_generate_uuid_unif_spread(self):
        n = 1000
        # union bound: (1/16)^32 * n^2 = 3e-33
        s = set([generate_uuid_unif() for i in xrange(n)])
        self.assertEquals(len(s), n)


class Duck(object):
    pass


class TestAnalyticUUIDGeneratorFactory(unittest.TestCase):

    def test_x_prefix_no_comm(self):
        n = 1000
        augf = AnalyticUUIDGeneratorFactory()
        u = augf.comm_uuid
        for i in xrange(n):
            aug = augf.create()
            self.assertTrue(aug.next().uuidString.startswith(u[:8 + 1 + 4]))

    def test_x_prefix_with_comm(self):
        n = 1000
        u = '7575a428-aaf7-4c2e-929e-1e2a0ab59e16'
        comm = Duck()
        comm.uuid = Duck()
        comm.uuid.uuidString = u
        augf = AnalyticUUIDGeneratorFactory(comm)
        self.assertEquals(augf.comm_uuid, u)
        for i in xrange(n):
            aug = augf.create()
            self.assertTrue(aug.next().uuidString.startswith(u[:8 + 1 + 4]))

    def test_x_prefix_bad_comm_uuid(self):
        u = '7575a428a-aaf7-4c2e-929e-1e2a0ab59e16'
        comm = Duck()
        comm.uuid = Duck()
        comm.uuid.uuidString = u
        augf = AnalyticUUIDGeneratorFactory(comm)
        with self.assertRaises(ValueError):
            augf.create()

    def test_y_prefix(self):
        m = 100
        n = 100
        augf = AnalyticUUIDGeneratorFactory()
        for i in xrange(m):
            aug = augf.create()
            uu = aug.next().uuidString
            for j in xrange(n - 1):
                self.assertTrue(aug.next().uuidString.startswith(
                    uu[:8 + 1 + 4 + 1 + 4 + 1 + 4]))

    def test_z_increment(self):
        m = 100
        n = 100
        augf = AnalyticUUIDGeneratorFactory()
        for i in xrange(m):
            aug = augf.create()
            u = aug.next().uuidString
            z = int(u[8 + 1 + 4 + 1 + 4 + 1 + 4 + 1:], 16)
            for j in xrange(n - 1):
                u = aug.next().uuidString
                self.assertEquals(
                    int(u[8 + 1 + 4 + 1 + 4 + 1 + 4 + 1:], 16),
                    (z + 1) % 2**48
                )
                z = (z + 1) % 2**48

    def test_x_prefix_spread(self):
        m = 100
        # union bound: (1/16)^12 * m^2 = 4e-11
        s = set()
        for i in xrange(m):
            augf = AnalyticUUIDGeneratorFactory()
            aug = augf.create()
            u = aug.next().uuidString
            s.add(u[:8 + 1 + 4])
        self.assertEquals(len(s), m)

    def test_y_prefix_spread(self):
        m = 10
        augf = AnalyticUUIDGeneratorFactory()
        # union bound: (1/16)^8 * m^2 = 2e-8
        s = set()
        for i in xrange(m):
            aug = augf.create()
            u = aug.next().uuidString
            s.add(u[:8 + 1 + 4 + 1 + 4 + 1 + 4])
        self.assertEquals(len(s), m)

    def test_spread(self):
        m = 100
        n = 100
        augf = AnalyticUUIDGeneratorFactory()
        # union bound: (2m-1)*(1/16)^12 * n^2 = 7e-9
        s = set()
        for i in xrange(m):
            aug = augf.create()
            u = aug.next().uuidString
            s.add(u)
            for j in xrange(n - 1):
                u = aug.next().uuidString
                s.add(u)
        self.assertEquals(len(s), m * n)



import logging

from concrete.structure.ttypes import TokenizationKind
from concrete.util.unnone import lun

from collections import deque


def get_tokens(tokenization, suppress_warnings=False):
    '''
    Return list of Tokens from lattice.cachedBestPath, if Tokenization
    kind is TOKEN_LATTICE; else, return list of Tokens from tokenList.

    Warn and return list of Tokens from tokenList if kind is not set.
    Return None if kind is set but the respective data fields are not.
    '''

    if tokenization.kind is None:
        if not suppress_warnings:
            logging.warn('Tokenization.kind is None but is required in '
                         'Concrete; using tokenList')

        token_list = tokenization.tokenList
        if token_list.tokenList is not None:
            return token_list.tokenList

    elif tokenization.kind == TokenizationKind.TOKEN_LATTICE:
        token_lattice = tokenization.lattice
        if token_lattice.cachedBestPath is not None:
            lattice_path = token_lattice.cachedBestPath
            if lattice_path.tokenList is not None:
                return lattice_path.tokenList

    elif tokenization.kind == TokenizationKind.TOKEN_LIST:
        token_list = tokenization.tokenList
        if token_list.tokenList is not None:
            return token_list.tokenList

    else:
        raise ValueError('Unrecognized TokenizationKind %s'
                         % tokenization.kind)

    return None


def get_tagged_tokens(tokenization, tagging_type):
    '''
    Return list of TaggedTokens of taggingType equal to tagging_type,
    if there is a unique choice.

    Raise exception if there is no matching tagging or more than one
    matching tagging.
    '''
    tts = [
        tt
        for tt in tokenization.tokenTaggingList
        if tt.taggingType == tagging_type
    ]
    if len(tts) == 0:
        raise Exception('No %s tagging.' % tagging_type)
    elif len(tts) == 1:
        return tts[0].taggedTokenList
    else:
        raise Exception('More than one %s tagging.' % tagging_type)


def get_lemmas(t):
    return get_tagged_tokens(t, 'LEMMA')


def get_pos(t):
    return get_tagged_tokens(t, 'POS')


def plus(x, y):
    return x + y


def flatten(a):
    return reduce(plus, a, [])


def get_comm_tokens(comm, sect_pred=None, suppress_warnings=False):
    '''
    Return list of Tokens in communication, delegating to get_tokens
    for each sentence.
    '''
    return flatten(map(
        lambda sect: flatten(map(
            lambda sent: get_tokens(sent.tokenization, suppress_warnings),
            sect.sentenceList
        )),
        filter(sect_pred, comm.sectionList)
        if (sect_pred is not None) else comm.sectionList
    ))


def get_comm_tokenizations(comm, tool=None):
    for section in lun(comm.sectionList):
        for sentence in lun(section.sentenceList):
            if tool is None or sentence.tokenization.metadata.tool == tool:
                yield sentence.tokenization


def _lattice_to_fsm(lattice):
    fsm = {}
    bkFsm = {}

    tokens = set()
    states = set()
    for arc in lattice.arcList:
        if arc.src is None:
            raise ValueError('Arc.src must be set')
        src = arc.src

        if arc.dst is None:
            raise ValueError('Arc.dst must be set')
        dst = arc.dst

        if arc.token is None:
            raise ValueError('Arc.token must be set')
        token = arc.token.tokenIndex

        if arc.weight is None:
            raise ValueError('Arc.weight must be set')
        wt = arc.weight

        tokens.add(token)
        states.add(src)
        states.add(dst)

        if src not in fsm:
            fsm[src] = {}
        if dst not in fsm[src]:
            fsm[src][dst] = []
        fsm[src][dst].append((token, wt))

        if dst not in bkFsm:
            bkFsm[dst] = {}
        if src not in bkFsm[dst]:
            bkFsm[dst][src] = []
        bkFsm[dst][src].append((token, wt))

    return (fsm, bkFsm, tokens, states)


def _calc_marginal_in_log_prob(fsm, states, start, end):
    '''
    Calculate marginal in-log-probability of each state.
    '''

    from scipy.misc import logsumexp

    alpha = {}
    alpha[start] = 0.

    state_queue = deque([start])

    while state_queue:
        currState = state_queue.popleft()

        if currState == end:
            continue

        arcs = fsm[currState]

        for (dst, tokenWts) in arcs.items():
            if dst not in alpha:
                state_queue.append(dst)
                alpha[dst] = float('-inf')
            for token, wt in tokenWts:
                alpha[dst] = logsumexp([alpha[dst], alpha[currState] + wt])

    return alpha


def compute_lattice_expected_counts(lattice):
    '''
    Given an object of type TokenLattice in which the
    dst, src, token, and weight fields are set in each arc,
    compute and return a list of expected token log-probabilities.
    Return type is a list of floats (expected log-probabilities)
    with the float at position i corresponding to the token with
    tokenIndex i.
    Input arc weights are treated as unnormalized log-probabilities.
    '''

    from scipy.misc import logsumexp

    (fsm, bkFsm, tokens, states) = _lattice_to_fsm(lattice)
    alpha = _calc_marginal_in_log_prob(fsm, states,
                                       lattice.startState, lattice.endState)
    beta = _calc_marginal_in_log_prob(bkFsm, states,
                                      lattice.endState, lattice.startState)
    norm = alpha[lattice.endState]

    expectedCounts = {}

    for state in states:
        currState = state

        if currState == lattice.endState:
            continue

        arcs = fsm[currState]

        for dst, tokenWts in arcs.items():
            for token, wt in tokenWts:
                if token not in expectedCounts:
                    expectedCounts[token] = []
                expectedCounts[token].append(
                    alpha[currState] + beta[dst] + wt - norm)

    if expectedCounts:
        return [
            (logsumexp(expectedCounts[idx]) if idx in expectedCounts else None)
            for idx in xrange(max(expectedCounts) + 1)
        ]
    else:
        return []



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.10.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
import sys
import concrete.metadata.ttypes
import concrete.language.ttypes
import concrete.structure.ttypes
import concrete.entities.ttypes
import concrete.situations.ttypes
import concrete.exceptions.ttypes
import concrete.email.ttypes
import concrete.twitter.ttypes
import concrete.audio.ttypes
import concrete.communication.ttypes

from thrift.transport import TTransport



#!/usr/bin/env python

'''
Convert Tweet file to Concrete Communications file.
'''

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from multiprocessing import Pool
import gzip
import json
import logging
import logging.config
import mimetypes

import concrete.version
from concrete.validate import validate_communication
from concrete.util.mem_io import write_communication_to_buffer
from concrete.util.twitter import json_tweet_string_to_Communication


def json_str_to_validated_concrete_bytes(tweet_str):
    tweet_str = tweet_str.decode('utf-8')
    b = json_tweet_string_to_Communication(tweet_str, True, True)
    if b is None or not validate_communication(b):
        return None
    else:
        return write_communication_to_buffer(b)


def json_str_to_concrete_bytes(tweet_str):
    tweet_str = tweet_str.decode('utf-8')
    b = json_tweet_string_to_Communication(tweet_str, True, True)
    if b is None:
        return None
    else:
        return write_communication_to_buffer(b)


def json_str_to_concrete_bytes_skip_bad_lines(tweet_str):
    try:
        tweet_str = tweet_str.decode('utf-8')
        return json_str_to_concrete_bytes(tweet_str)
    except:
        return None


def json_str_to_validated_concrete_bytes_skip_bad_lines(tweet_str):
    try:
        tweet_str = tweet_str.decode('utf-8')
        return json_str_to_concrete_bytes(tweet_str)
    except:
        return None


def main():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description='Read tweets formatted in the Twitter JSON API and write'
                    ' communications',
    )
    parser.set_defaults(log_level='INFO', num_proc=1, chunk_size=10000,
                        log_interval=10000)
    parser.add_argument('tweet_path', type=str,
                        help='Input twitter JSON file path (- for stdin)')
    parser.add_argument('concrete_path', type=str,
                        help='Output concrete file path (- for stdout)')
    parser.add_argument('--log-interval', type=int,
                        help='log an info message every log-interval tweets')
    parser.add_argument('--log-level', type=str,
                        choices=('DEBUG', 'INFO', 'WARNING', 'ERROR'),
                        help='Logging verbosity level (to stderr)')
    parser.add_argument('--log-conf-path', type=str,
                        help='Path to log config file (overrides log-level).'
                             ' Format is json encoding of logging.config'
                             ' dictConfig schema: https://docs.python.org/2/'
                             'library/logging.config.html#'
                             'logging-config-dictschema')
    parser.add_argument('--num-proc', type=int,
                        help='Number of worker processes to use')
    parser.add_argument('--chunk-size', type=int,
                        help='Chunk size (in number of tweets) when'
                             ' dispatching jobs to workers')
    parser.add_argument('--catch-ioerror', action='store_true',
                        help='Treat IOError as successful end of stream'
                             ' (recommended if gzipped input files were not'
                             ' completely written)')
    parser.add_argument('--skip-bad-lines', action='store_true',
                        help='Skip malformatted json lines'
                             ' (by default, they crash the program)')
    parser.add_argument('--skip-invalid-comms', action='store_true',
                        help='Skip invalid communications (increases runtime)'
                             ' (by default, validation is not performed)')
    concrete.version.add_argparse_argument(parser)
    ns = parser.parse_args()

    # Won't work on Windows... but that use case is very unlikely
    tweet_path = '/dev/fd/0' if ns.tweet_path == '-' else ns.tweet_path
    concrete_path = (
        '/dev/fd/1' if ns.concrete_path == '-' else ns.concrete_path
    )

    if ns.log_conf_path:
        with open(ns.log_conf_path) as f:
            logging.config.dictConfig(json.load(f))
    else:
        logging.basicConfig(
            level=ns.log_level,
            format='%(asctime)-15s %(levelname)s: %(message)s'
        )

    if ns.tweet_path != '-' and mimetypes.guess_type(tweet_path)[1] == 'gzip':
        tweet_reader = gzip.open(tweet_path, 'r')
    else:
        tweet_reader = open(tweet_path, 'rb')

    if ns.catch_ioerror:
        def _catch(g):
            it = iter(g)
            while True:
                try:
                    x = it.next()
                    yield x
                except IOError:
                    raise StopIteration('Caught IOError.')
        tweet_reader = _catch(tweet_reader)

    if ns.skip_bad_lines:
        if ns.skip_invalid_comms:
            map_func = json_str_to_validated_concrete_bytes_skip_bad_lines
        else:
            map_func = json_str_to_concrete_bytes_skip_bad_lines
    else:
        if ns.skip_invalid_comms:
            map_func = json_str_to_validated_concrete_bytes
        else:
            map_func = json_str_to_concrete_bytes

    with open(concrete_path, 'wb') as writer:
        i = 0

        p = Pool(ns.num_proc)
        for concrete_bytes in p.imap(map_func, tweet_reader,
                                     ns.chunk_size):
            if concrete_bytes is not None:
                if (i + 1) % ns.log_interval == 0:
                    logging.info('writing tweet %d...' % (i + 1))
                writer.write(concrete_bytes)
                i += 1


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""Command line client for FetchCommunicationService
"""

import argparse

import concrete.version
from concrete.access import FetchCommunicationService
from concrete.access.ttypes import FetchRequest
from concrete.util.file_io import write_communication_to_file
from concrete.util.thrift_factory import factory


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="Command line client for interacting with a FetchCommunicationService server",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("-p", "--port", type=int, default=9090,
                        help="Port of FetchCommunicationService server")
    parser.add_argument("-s", "--server", default="localhost",
                        help="Hostname of FetchCommunicationService server")
    parser.add_argument("--about", action="store_true",
                        help="Print value of fetch_service.about()")
    parser.add_argument("--alive", action="store_true",
                        help="Print value of fetch_service.alive()")
    parser.add_argument("--count", action="store_true",
                        help="Print value of fetch_service.getCommunicationCount()")
    parser.add_argument("--get-ids", action="store_true",
                        help="Print list of Communication IDs returned by "
                        "fetch_service.getCommunicationIDs(offset, count).  "
                        "The offset and count parameters are set using the "
                        "'--get-ids-offset' and '--get-ids-count' flags")
    parser.add_argument("--get-ids-offset", type=int, default=0,
                        help="Number of Communication IDs printed using the '--get-ids' flag")
    parser.add_argument("--get-ids-count", type=int, default=20,
                        help="Offset for Communication IDs printed using the '--get-ids' flag")
    parser.add_argument("--save", action="store_true",
                        help="Save fetched Communications to disk as "
                        "'[COMMUNICATION_ID].concrete'")
    parser.add_argument("comm_id", nargs="*", help="IDs of Communications to be fetched")
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    socket = factory.createSocket(args.server, args.port)
    transport = factory.createTransport(socket)
    protocol = factory.createProtocol(transport)
    client = FetchCommunicationService.Client(protocol)
    transport.open()

    if args.comm_id:
        fetch_request = FetchRequest()
        fetch_request.communicationIds = args.comm_id
        fetch_result = client.fetch(fetch_request)
        print "Received FetchResult: '%s'" % fetch_result

    if args.about:
        print "FetchCommunicationService.about() returned %s" % client.about()
    if args.alive:
        print "FetchCommunicationService.alive() returned %s" % client.alive()
    if args.count:
        print "FetchCommunicationService.getCommunicationCount() returned %d" % \
            client.getCommunicationCount()
    if args.get_ids:
        print "FetchCommunicationService.getCommunicationIDs(offset=%d, count=%d) returned:" % \
            (args.get_ids_offset, args.get_ids_count)
        for comm_id in client.getCommunicationIDs(args.get_ids_offset, args.get_ids_count):
            print "  %s" % comm_id

    if args.save:
        if fetch_result.communications:
            for comm in fetch_result.communications:
                comm_filename = '%s.concrete' % comm.id
                print "Saving Communication as '%s'" % comm_filename
                write_communication_to_file(comm, comm_filename)



import logging

from concrete.access.ttypes import FetchResult
from concrete.services.ttypes import ServiceInfo
from concrete.util.access_wrapper import FetchCommunicationClientWrapper
import concrete.version


class CommunicationContainerFetchHandler(object):
    """FetchCommunicationService implementation using Communication containers

    Implements the FetchCommunicationService interface, retrieving
    Communications from a dict-like `communication_container` object
    that maps Communication ID strings to Communications.  The
    `communication_container` could be an actual dict, or a container
    such as:
    - DirectoryBackedCommunicationContainer
    - FetchBackedCommunicationContainer
    - MemoryBackedCommunicationContainer
    - ZipFileBackedCommunicationContainer

    Usage:
        from concrete.util.access_wrapper import FetchCommunicationServiceWrapper

        handler = CommunicationContainerFetchHandler(comm_container)
        fetch_service = FetchCommunicationServiceWrapper(handler)
        fetch_service.serve(host, port)
    """

    def __init__(self, communication_container):
        """
        Args:
        - `communication_container`: Dict-like object that maps Communication
                                     IDs to Communications
        """
        self.communication_container = communication_container

    def about(self):
        logging.info("Received about() call")
        service_info = ServiceInfo()
        service_info.name = 'CommunicationContainerFetchHandler - %s' % \
                            type(self.communication_container)
        service_info.version = concrete.version.concrete_library_version()
        return service_info

    def alive(self):
        logging.info("Received alive() call")
        return True

    def fetch(self, fetch_request):
        logging.info("Received FetchRequest: %s" % fetch_request)
        fetch_result = FetchResult()
        fetch_result.communications = []
        for communication_id in fetch_request.communicationIds:
            if communication_id in self.communication_container:
                comm = self.communication_container[communication_id]
                fetch_result.communications.append(comm)
            else:
                logging.warning('Unable to find Communication with ID: %s' % communication_id)
        return fetch_result

    def getCommunicationCount(self):
        logging.info('Received getCommunicationCount()')
        communicationCount = len(self.communication_container)
        logging.info('- Communication Count: %d' % communicationCount)
        return communicationCount

    def getCommunicationIDs(self, offset, count):
        logging.info('Received getCommunicationIDs() call')
        return self.communication_container.keys()[offset:][:count]


class RelayFetchHandler(object):
    """Implements a 'relay' to another FetchCommunicationService server.

    A FetchCommunicationService that acts as a relay to a second
    FetchCommunicationService, where the second service is using the
    TSocket transport and TCompactProtocol protocol.

    This class was designed for the use case where you have Thrift
    JavaScript code that needs to communicate with a
    FetchCommunicationService server, but the server does not support
    the same Thrift serialization protocol as the JavaScript client.

    The de-facto standard for Concrete services is to use the
    TCompactProtocol serialization protocol over a TSocket connection.
    But as of Thrift 0.10.0, the Thrift JavaScript libraries only
    support using TJSONProtocol over HTTP.

    The RelayFetchHandler class is intended to be used as server-side
    code by a web application.  The JavaScript code will make
    FetchCommunicationService RPC calls to the web server using
    HTTP/TJSONProtocol, and the web application will then pass these
    RPC calls to another FetchCommunicationService using
    TSocket/TCompactProtocol RPC calls.
    """
    def __init__(self, host, port):
        """
        Args:
        - `host`: Hostname of FetchCommunicationService server
        - `port`: Port # of FetchCommunicationService server
        """
        self.host = host
        self.port = int(port)

    def about(self):
        logging.info('RelayFetchHandler.about()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.about()

    def alive(self):
        logging.info('RelayFetchHandler.alive()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.alive()

    def fetch(self, request):
        logging.info('RelayFetchHandler.fetch()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.fetch(request)

    def getCommunicationCount(self):
        logging.info('RelayFetchHandler.getCommunicationCount()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.getCommunicationCount()

    def getCommunicationIDs(self, offset, count):
        logging.info('RelayFetchHandler.getCommunicationIDs(offset=%d, count=%d)' %
                     (offset, count))
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.getCommunicationIDs(offset, count)



"""Communication Containers - mapping Communication IDs to Communications

Classes that behave like a read-only dictionary (implementing Python's
collections.Mapping interface) and map Communication ID strings to
Communications.

The classes abstract away the storage backend.  If you need to
optimize for performance, you may not want to use a dictionary
abstraction that retrieves one Communication at a time.
"""

import collections
import gzip
import logging
import os
import zipfile

import humanfriendly

import concrete
from concrete.access.ttypes import FetchRequest
from concrete.util.access_wrapper import FetchCommunicationClientWrapper
from concrete.util.file_io import (
    CommunicationReader,
    read_communication_from_file)
from concrete.util.mem_io import read_communication_from_buffer


class DirectoryBackedCommunicationContainer(collections.Mapping):
    """Maps Comm IDs to Comms, retrieving Comms from the filesystem

    DirectoryBackedCommunicationContainer instances behave as
    dict-like data structures that map Communication IDs to
    Communications.  Communications are lazily-retrieved from the
    filesystem.

    Upon initialization, a DirectoryBackedCommunicationContainer
    instance will (recursively) search `directory_path` for any files
    that end with the specified `comm_extensions`.  Files with
    matching extensions are assumed to be Communication files whose
    filename (sans extension) is the file's Communication ID.  So, for
    example, a file named 'XIN_ENG_20101212.0120.concrete' is assumed
    to be a Communication file with a Communication ID of
    'XIN_ENG_20101212.0120'.

    Files with the extension `.gz` will be decompressed using gzip.

    A DirectoryBackedCommunicationsContainer will not be able to find
    any files that are added to `directory_path` after the container
    was initialized.
    """

    def __init__(self, directory_path,
                 comm_extensions=['.comm', '.concrete', '.gz']):
        """
        Args:
        - `directory_path`: Path to directory containing Communications files
        - `comm_extensions`: List of strings specifying filename extensions
                             associated with Communications
        """
        self.comm_id_to_comm_path = {}

        logging.info("Caching names of files with extensions [%s] in '%s'" %
                     (', '.join(comm_extensions), directory_path))

        # Recursively traverse subdirectories
        for root, dirs, files in os.walk(directory_path):
            for basename in files:
                (comm_id, extension) = os.path.splitext(basename)
                if extension in comm_extensions:
                    comm_filename = os.path.join(root, basename)
                    self.comm_id_to_comm_path[comm_id] = comm_filename

        logging.info("Finished caching %d Communication filenames in '%s'" %
                     (len(self.comm_id_to_comm_path), directory_path))

    def __getitem__(self, communication_id):
        if communication_id in self.comm_id_to_comm_path:
            comm_path = self.comm_id_to_comm_path[communication_id]
            logging.info("Reading Communication with ID '%s'" % comm_path)
            if not os.path.exists(comm_path):
                logging.error("Unable to find file with path '%s'" % comm_path)
                raise KeyError
            if os.path.splitext(comm_path)[1] == '.gz':
                with gzip.open(comm_path) as gzip_file:
                    buf = gzip_file.read()
                    comm = read_communication_from_buffer(buf)
            else:
                comm = read_communication_from_file(comm_path)
            return comm
        else:
            logging.info('No Communication with ID: %s' % communication_id)
            raise KeyError

    def __iter__(self):
        return self.comm_id_to_comm_path.__iter__()

    def __len__(self):
        return len(self.comm_id_to_comm_path)


class FetchBackedCommunicationContainer(collections.Mapping):
    """Maps Comm IDs to Comms, retrieving Comms from a Fetch server

    FetchBackedCommunicationContainer instances behave as dict-like data
    structures that map Communication IDs to Communications.  Communications
    are lazily-retrieved from a FetchCommunicationService.

    If you need to retrieve large amounts of data from a
    FetchCommunicationService, then you SHOULD NOT USE THIS CLASS.
    This class retrieves one Communication at a time using
    FetchCommunicationService.
    """

    def __init__(self, host, port):
        """
        Args:
        - `host`: Hostname of FetchCommunicationService server
        - `port`: Port # of FetchCommunicationService server
        """
        self.host = host
        self.port = port

    def __getitem__(self, communication_id):
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            fetch_result = fc.fetch(FetchRequest(communicationIds=[communication_id]))
            total_results = len(fetch_result.communications)
            if total_results == 0:
                raise KeyError
            elif total_results == 1:
                return fetch_result.communications[0]
            else:
                raise Exception("FetchBackedCommunicationContainer.__get_item__() "
                                "expected to receive 1 Communication, but instead "
                                "received %d Communications" % total_results)

    def __iter__(self):
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            n = fc.getCommunicationCount()
            return fc.getCommunicationIDs(0, n).__iter__()

    def __len__(self):
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.getCommunicationCount()


class MemoryBackedCommunicationContainer(collections.Mapping):
    """Maps Comm IDs to Comms by loading all Comms in file into memory

    FetchBackedCommunicationContainer instances behave as dict-like
    data structures that map Communication IDs to Communications.  All
    Communications in `communications_file` will be read into memory
    using a CommunicationReader instance.
    """

    def __init__(self, communications_file, max_file_size=1073741824):
        """
        Args:
        `communications_file`: String specifying name of Communications file
        `max_file_size`: Integer specifying maximum file size, in bytes
        """
        self.comm_id_to_comm = {}

        comm_file_size = os.path.getsize(communications_file)
        if comm_file_size > max_file_size:
            raise Exception(
                ("MemoryBackedComunicationContainer will not open the file "
                 "'%s' because the file's size (%s) is larger than the "
                 "maximum specified file size of %s.  If you would like to "
                 "read a file this large into memory, you will need to "
                 "specify a larger value for the 'max_file_size'parameter.") %
                (communications_file,
                 humanfriendly.format_size(comm_file_size, binary=True),
                 humanfriendly.format_size(max_file_size, binary=True)))

        logging.info("Reading in Communications from file '%s'" %
                     communications_file)
        logging.info("Communication IDs:")
        for (comm, _) in CommunicationReader(communications_file):
            self.comm_id_to_comm[comm.id] = comm
            logging.info("  %s" % comm.id)
        logging.info("Finished reading communications.\n")

    def __getitem__(self, communication_id):
        return self.comm_id_to_comm[communication_id]

    def __iter__(self):
        return self.comm_id_to_comm.__iter__()

    def __len__(self):
        return len(self.comm_id_to_comm)


class ZipFileBackedCommunicationContainer(collections.Mapping):
    """Maps Comm IDs to Comms, retrieving Comms from a Zip file

    ZipFileBackedCommunicationContainer instances behave as dict-like
    data structures that map Communication IDs to Communications.
    Communications are lazily-retrieved from a Zip file.
    """

    def __init__(self, zipfile_path, comm_extensions=['.comm', '.concrete']):
        """
        Args:
        - `zipfile_path`: Path to Zip file containing Communications
        - `comm_extensions`: List of strings specifying filename extensions
                             associated with Communications
        """
        self.comm_id_to_filename = {}
        self.zipfile = zipfile.ZipFile(zipfile_path, 'r')
        for filename in self.zipfile.namelist():
            (comm_id, extension) = os.path.splitext(os.path.basename(filename))
            if extension in comm_extensions:
                self.comm_id_to_filename[comm_id] = filename

    def __exit__(self):
        self.zipfile.close()

    def __getitem__(self, communication_id):
        filename = self.comm_id_to_filename[communication_id]
        buf = self.zipfile.read(filename)
        comm = concrete.util.read_communication_from_buffer(buf)
        return comm

    def __iter__(self):
        return self.comm_id_to_filename.__iter__()

    def __len__(self):
        return len(self.comm_id_to_filename)



#!/usr/bin/env python

import concrete.version
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from concrete.util import CommunicationReader, CommunicationWriter, FileType
from concrete.util.annotate_wrapper import AnnotateCommunicationClientWrapper


def main():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description="Interface with a Concrete AnnotateCommunicationService"
    )
    parser.add_argument('host',
                        help="Hostname of annotate service to which to"
                             " connect.")
    parser.add_argument('port', type=int,
                        help="Port of annotate service to which to connect.")
    parser.add_argument('--input', default='-',
                        help="Input source to use. '-' for stdin; otherwise"
                             " takes a path to a file.")
    parser.add_argument('--output', default='-',
                        help="Output source to use. '-' for stdout; otherwise"
                             " takes a path to a file.")
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    # Won't work on Windows... but that use case is very unlikely
    if args.input == '-':
        reader_kwargs = dict(filetype=FileType.STREAM)
        input_path = '/dev/fd/0'
    else:
        reader_kwargs = dict()
        input_path = args.input
    output_path = '/dev/fd/1' if args.output == '-' else args.output

    reader = CommunicationReader(input_path, **reader_kwargs)
    with AnnotateCommunicationClientWrapper(args.host, args.port) as client:
        with CommunicationWriter(output_path) as writer:
            for (comm, _) in reader:
                writer.write(client.annotate(comm))


if __name__ == "__main__":
    main()



"""
Helper functions for generating Concrete UUIDs
"""

# Force 'import uuid' to import the Python standard library module
# named "uuid", and not the "concrete.uuid" module
from __future__ import absolute_import

import uuid as python_uuid

from thrift.Thrift import TType
from concrete.uuid.ttypes import UUID
from concrete.metadata.ttypes import AnnotationMetadata
from concrete.util.mem_io import communication_deep_copy

from inspect import isroutine
import random
import logging


def generate_UUID():
    """Helper function for generating a Concrete UUID object
    """
    return UUID(uuidString=str(python_uuid.uuid4()))


def hex_to_bin(h):
    return int(h, 16)


def bin_to_hex(b, n=None):
    h = hex(b)[2:]
    if h.endswith('L'):
        h = h[:-1]
    if n is None:
        n = len(h)
    elif len(h) > n:
        raise ValueError('hex string "%s" is longer than %d chars' % (h, n))
    return ('0' * (n - len(h))) + h


def split_uuid(u):
    p = u.split('-')

    valid_input = (
        len(p) == 5 and
        len(p[0]) == 8 and
        len(p[1]) == 4 and
        len(p[2]) == 4 and
        len(p[3]) == 4 and
        len(p[4]) == 12
    )
    if not valid_input:
        raise ValueError(
            'uuid "%s" is not of the form xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'
            % u
        )

    xs = p[0] + p[1]
    ys = p[2] + p[3]
    zs = p[4]

    return (xs, ys, zs)


def join_uuid(xs, ys, zs):
    valid_input = (
        len(xs) == 12 and
        len(ys) == 8 and
        len(zs) == 12
    )
    if not valid_input:
        raise ValueError('uuid pieces do not have lengths 12, 8, 12')

    u = '-'.join((xs[:8], xs[8:], ys[:4], ys[4:], zs))

    return u


def generate_hex_unif(n):
    return ''.join(random.choice('abcdef0123456789') for i in xrange(n))


def generate_uuid_unif():
    return join_uuid(generate_hex_unif(12),
                     generate_hex_unif(8),
                     generate_hex_unif(12))


class _AnalyticUUIDGenerator(object):
    '''
    UUID generator for a given analytic in a given communication.
    '''

    def __init__(self, u):
        (self._xs, ys, zs) = split_uuid(u)
        self._ys = generate_hex_unif(len(ys))
        self._z = hex_to_bin(generate_hex_unif(len(zs)))
        self._z_len = len(zs)
        self._z_bound = 2**(4 * len(zs))
        self.n = 0

    def __iter__(self):
        return self

    def next(self):
        '''
        Generate and return a new concrete UUID.
        StopIteration will never be raised.
        '''
        self._z = (self._z + 1) % self._z_bound
        self.n += 1
        return UUID(uuidString=join_uuid(
            self._xs, self._ys, bin_to_hex(self._z, self._z_len)
        ))


class AnalyticUUIDGeneratorFactory(object):
    '''
    Factory for a compressible UUID generator.

    One factory should be created per communication, and a new generator
    should be created from that factory for each analytic processing the
    communication.  Usually each program represents a single analytic,
    so common usage is

        augf = AnalyticUUIDGeneratorFactory(comm)
        aug = augf.create()
        for <each annotation object created by this analytic>:
            annotation. = aug.next()
            <add annotation to communication>

    or if you're creating a new communication

        augf = AnalyticUUIDGeneratorFactory()
        aug = augf.create()
        comm = <create communication>
        comm.uuid = aug.next()
        for <each annotation object created by this analytic>:
            annotation. = aug.next()
            <add annotation to communication>

    where the annotation objects might be objects of type
    Parse, DependencyParse, TokenTagging, CommunicationTagging, etc.
    '''

    def __init__(self, comm=None):
        if comm is None:
            self.comm_uuid = generate_uuid_unif()
        else:
            self.comm_uuid = comm.uuid.uuidString

    def create(self):
        '''
        Create and return a UUID generator for a new analytic.
        '''
        return _AnalyticUUIDGenerator(self.comm_uuid)


def _filtered_getmembers(obj):
    '''
    Generate key-value pairs of object members that may contain UUIDs.
    Over-generate, but filter the output enough that concrete objects
    can be traversed recursively using this function without leading to
    stack overflows or infinite loops.
    '''

    for k in dir(obj):
        if not (k[0] == '_' or k == 'thrift_spec' or k == 'read' or
                k == 'write' or k == 'validate'):
            v = getattr(obj, k)
            if not (isroutine(v) or
                    isinstance(v, int) or isinstance(v, float) or
                    isinstance(v, str) or isinstance(v, unicode)):
                yield (k, v)


_FILTERED_TTYPES = set((TType.STRUCT, TType.LIST, TType.MAP, TType.SET))


def _fast_filtered_getmembers(obj):
    'Fast thrift-specific implementation of filtered_getmembers.'

    if hasattr(obj, 'thrift_spec'):
        for s in obj.thrift_spec:
            if s is not None:
                t = s[1]
                if t in _FILTERED_TTYPES:
                    k = s[2]
                    yield (k, getattr(obj, k))


class UUIDClustering(object):
    '''
    Representation of the UUID instance clusters in a concrete
    communication (each cluster represents the set of nested members of
    the communication that reference or are identified by a given UUID).
    '''

    def __init__(self, comm):
        self._clusters = dict()  # map: UUID -> set of nested members
        self._search(comm)

    def hashable_clusters(self):
        '''
        Return the set of unlabeled UUID clusters in a unique and
        hashable format.  Two UUIDClusterings c1 and c2 are equivalent
        (the two underlying communications' UUID structures are
        equivalent) if and only if:

            c1.hashable_clusters() == c2.hashable_clusters()
        '''
        return set(tuple(sorted(c)) for c in self._clusters.values())

    def _search(self, obj, prefix=()):
        '''
        Search obj for UUIDs, calling _add_uuid_field when UUIDs are
        found and calling _search on other object members.
        When _search calls itself, prefix is appended with the object
        member name, forming a uniquely identifiable tuple
        representation of the path from the root object to a nested
        object member.
        '''

        if isinstance(obj, UUID):
            self._add_uuid_field(obj.uuidString, prefix)
        elif isinstance(obj, list):
            for (i, v) in enumerate(obj):
                self._search(v, prefix + (('list', i),))
        elif isinstance(obj, set):
            raise ValueError('UUIDClustering does not support sets')
        elif isinstance(obj, dict):
            for (k, v) in obj.items():
                self._search(v, prefix + (('dict', k),))
        else:
            for (k, v) in _filtered_getmembers(obj):
                self._search(v, prefix + (k,))

    def _add_uuid_field(self, u, f):
        '''
        Add UUID field f (a unique, hashable representation of the path
        from the root communication to a nested UUID object) to the UUID
        cluster indexed by UUID string u.
        '''
        if u in self._clusters:
            self._clusters[u].add(f)
        else:
            self._clusters[u] = set([f])


class UUIDCompressor(object):

    def __init__(self, single_analytic=False):
        self.single_analytic = single_analytic

    def compress(self, comm):
        'Return a deep copy of comm with compressed UUIDs.'

        cc = communication_deep_copy(comm)
        self.augf = AnalyticUUIDGeneratorFactory(cc)
        self.augs = dict()
        self.uuid_map = dict()

        self._compress_uuids(cc)
        self._compress_uuid_refs(cc)

        return cc

    def _compress_uuids(self, obj, name_is_uuid=False, tool=None):
        'Generate new UUIDs in "uuid" fields and save mapping'

        tool = self._get_tool(obj, tool)

        if name_is_uuid:
            if isinstance(obj, UUID):
                obj.uuidString = self._gen_uuid(obj, tool)
            else:
                logging.warning('uuid not instance of UUID')

        if not isinstance(obj, UUID):  # we already took care of "uuid"
            self._apply(
                lambda elt, elt_name_is_uuid: self._compress_uuids(
                    elt, name_is_uuid=elt_name_is_uuid, tool=tool
                ),
                obj
            )

    def _compress_uuid_refs(self, obj, name_is_uuid=False, tool=None):
        'Update UUID references (not in "uuid" fields) using saved mapping'

        tool = self._get_tool(obj, tool)

        if isinstance(obj, UUID):
            if not name_is_uuid:
                obj.uuidString = self.uuid_map[obj.uuidString]
        else:
            self._apply(
                lambda elt, elt_name_is_uuid: self._compress_uuid_refs(
                    elt, name_is_uuid=elt_name_is_uuid, tool=tool
                ),
                obj
            )

    def _get_tool(self, obj, tool=None):
        '''
        Return tool for this object, given the parent tool;
        update self.augs
        '''

        if hasattr(obj, 'metadata'):
            if isinstance(obj.metadata, AnnotationMetadata):
                tool = obj.metadata.tool
            else:
                logging.warning('metadata not instance of AnnotationMetadata')
        if self.single_analytic:
            tool = None
        if tool not in self.augs:
            self.augs[tool] = self.augf.create()
        return tool

    def _gen_uuid(self, old_uuid, tool):
        '''
        Return a new UUID for the provided tool, using self.augs;
        update self.uuid_map
        '''

        aug = self.augs[tool]
        new_uuid = aug.next()
        if old_uuid.uuidString in self.uuid_map:
            raise ValueError('encountered UUID %s twice, aborting' %
                             old_uuid.uuidString)
        self.uuid_map[old_uuid.uuidString] = new_uuid.uuidString
        return new_uuid.uuidString

    @classmethod
    def _apply(cls, f, x):
        '''
        Apply f to the members of x if it is a basic container type,
        otherwise apply f to x directly.
        '''

        if isinstance(x, list):
            for elt in x:
                f(elt, False)
        elif isinstance(x, set):
            for elt in x:
                f(elt, False)
        elif isinstance(x, dict):
            for elt in x.values():
                f(elt, False)
        else:
            for (k, v) in _fast_filtered_getmembers(x):
                f(v, k == 'uuid')


def compress_uuids(comm, verify=False, single_analytic=False):
    '''
    Create a copy of communication comm with UUIDs converted according
    to the compressible UUID scheme.  Return a 2-tuple containing that
    new communication and the UUIDCompressor object used to perform
    the conversion.

    If verify is True, use a heuristic to verify the UUID link structure
    is preserved in the new communication.  If single_analytic is True,
    use a single analytic prefix for all UUIDs in comm.

    If verify is True and comm has references added, throw a ValueError
    because verification would cause an infinite loop.
    '''

    if verify and hasattr(comm, 'tokenizationForUUID'):
        raise ValueError('cannot verify communication with references')

    uc = UUIDCompressor(single_analytic=single_analytic)

    new_comm = uc.compress(comm)

    num_old_uuids = len(set(uc.uuid_map.keys()))
    num_new_uuids = len(set(uc.uuid_map.values()))

    if verify:
        c1 = UUIDClustering(comm).hashable_clusters()
        c2 = UUIDClustering(new_comm).hashable_clusters()

        # Verification is c1 == c2;
        # also check UUID map lengths are the same as a sanity-check
        if num_old_uuids == num_new_uuids and c1 == c2:
            logging.info('verified %s (%d uuid instances, %d uuids)'
                         % (comm.id, sum(len(c) for c in c1), len(c1)))
        else:
            logging.error('%s failed verification' % comm.id)
            logging.error('uuid counts: %d -> %d'
                          % (num_old_uuids, num_new_uuids))
            logging.error('verified number of uuids: %d -> %d'
                          % (len(c1), len(c2)))
            logging.error('verified number of uuid instances: %d -> %d'
                          % (sum(map(len, c1)), sum(map(len, c2))))
            raise Exception('%s failed verification' % comm.id)

    else:
        if num_old_uuids != num_new_uuids:
            logging.warning('uuid counts are not the same (%d -> %d)'
                            % (num_old_uuids, num_new_uuids))

    return (new_comm, uc)



from datetime import datetime


EPOCH = datetime.utcfromtimestamp(0)


def datetime_to_timestamp(dt):
    '''
    Source:
    http://stackoverflow.com/questions/6999726/how-can-i-convert-a-datetime-object-to-milliseconds-since-epoch-unix-time-in-p
    '''
    return int((dt - EPOCH).total_seconds())


def get_index_of_tool(lst_of_conc, tool):
    """Return the index of the object in the provided list
    whose tool name matches tool.

    If tool is None, return the first valid index into `lst_of_conc`.

    This returns -1 if:
      * `lst_of_conc` is None, or
      * `lst_of_conc` has no entries, or
      * no object in `lst_of_conc` matches `tool`.

    Args:

    - `lst_of_conc`: A list of Concrete objects, each of which
      has a `.metadata` field.
    - `tool`: A tool name to match.
    """
    idx = -1
    if lst_of_conc is not None and len(lst_of_conc) > 0:
        if tool is not None:
            for (cidx, obj) in enumerate(lst_of_conc):
                if obj.metadata.tool == tool:
                    idx = cidx
                    break
        else:
            idx = 0
    return idx



"""Functions used by `concrete_inspect.py` to print data in a Communication.

The function implementations provide useful examples of how to
interact with many different Concrete datastructures.
"""

from concrete.util.metadata import get_index_of_tool
from concrete.util.unnone import lun
from collections import defaultdict
from operator import attrgetter


def _reconcile_index_and_tool(lst_of_conc, given_idx, tool):
    """Given a list of Concrete objects with metadata (e.g., `DependencyParse`s)
    and a default index, find the index of the object whose `.metadata.tool`
    matches the provided query tool name `tool`.

    When no tool is provided (but with an iterable list), this returns the
    provided default index (even if it\'s not a valid index).
    If the tool isn't found, the list is None or empty, this returns -1.
    """
    valid_lst = lst_of_conc is not None and len(lst_of_conc) > 0
    idx = given_idx if valid_lst else -1
    if tool is not None:
        idx = get_index_of_tool(lst_of_conc, tool)
    return idx


def _valid_index_lun(lst, idx):
    """Return True iff `idx` is a valid index into
    the given (non-None) list. If `lst` is None,
    return False.
    """

    if lst is None or len(lst) == 0:
        return False
    return idx >= 0 and idx < len(lst)


def _filter_by_tool(lst, tool):
    return filter(lambda x: tool is None or x.metadata.tool == tool, lst)


def print_conll_style_tags_for_communication(
        comm, char_offsets=False, dependency=False, lemmas=False, ner=False,
        pos=False,
        dependency_tool=None, lemmas_tool=None, ner_tool=None, pos_tool=None):

    """Print 'ConLL-style' tags for the tokens in a Communication

    Args:

    - `comm`: A Concrete Communication object
    - `char_offsets`: A boolean flag for printing token text specified by
          a Token's (optional) TextSpan
    - `dependency`: A boolean flag for printing dependency parse HEAD tags
    - `lemmas`: A boolean flag for printing lemma tags
    - `ner`: A boolean flag for printing Named Entity Recognition tags
    - `pos`: A boolean flag for printing Part-of-Speech tags
    """

    header_fields = [u"INDEX", u"TOKEN"]
    if char_offsets:
        header_fields.append(u"CHAR")
    if lemmas:
        header_fields.append(u"LEMMA")
    if pos:
        header_fields.append(u"POS")
    if ner:
        header_fields.append(u"NER")
    if dependency:
        header_fields.append(u"HEAD")
        header_fields.append(u"DEPREL")
    print u"\t".join(header_fields)
    dashes = ["-" * len(fieldname) for fieldname in header_fields]
    print u"\t".join(dashes)

    for tokenization in _get_tokenizations(comm):
        token_tag_lists = []

        if char_offsets:
            token_tag_lists.append(
                _get_char_offset_tags_for_tokenization(comm, tokenization))
        if lemmas:
            token_tag_lists.append(
                _get_lemma_tags_for_tokenization(tokenization,
                                                 tool=lemmas_tool))
        if pos:
            token_tag_lists.append(
                _get_pos_tags_for_tokenization(tokenization, tool=pos_tool))
        if ner:
            token_tag_lists.append(
                _get_ner_tags_for_tokenization(tokenization, tool=ner_tool))
        if dependency:
            token_tag_lists.append(
                _get_conll_head_tags_for_tokenization(tokenization,
                                                      tool=dependency_tool))
            token_tag_lists.append(
                _get_conll_deprel_tags_for_tokenization(tokenization,
                                                        tool=dependency_tool))
        print_conll_style_tags_for_tokenization(tokenization,
                                                token_tag_lists)
        print


def print_conll_style_tags_for_tokenization(tokenization, token_tag_lists):
    """Print 'ConLL-style' tags for the tokens in a tokenization

    Args:

    - `tokenization`: A Concrete Tokenization object
    - `token_tag_lists`: A list of lists of token tag strings
    """
    if tokenization.tokenList:
        for i, token in enumerate(tokenization.tokenList.tokenList):
            token_tags = [unicode(token_tag_list[i])
                          for token_tag_list in token_tag_lists]
            fields = [unicode(i + 1), token.text]
            fields.extend(token_tags)
            print u"\t".join(fields)


def print_entities(comm, tool=None):
    """Print information for all Entities and their EntityMentions

    Args:

    - `comm`: A Concrete Communication object
    """
    if comm.entitySetList:
        for entitySet_index, entitySet in enumerate(comm.entitySetList):
            if tool is None or entitySet.metadata.tool == tool:
                print u"Entity Set %d (%s):" % (entitySet_index,
                                                entitySet.metadata.tool)
                for entity_index, entity in enumerate(entitySet.entityList):
                    print u"  Entity %d-%d:" % (entitySet_index, entity_index)
                    for em_index, em in enumerate(entity.mentionList):
                        print u"      EntityMention %d-%d-%d:" % (
                            entitySet_index, entity_index, em_index)
                        print u"          tokens:     %s" % (
                            u" ".join(_get_tokens_for_entityMention(em)))
                        if em.text:
                            print u"          text:       %s" % em.text
                        print u"          entityType: %s" % em.entityType
                        print u"          phraseType: %s" % em.phraseType
                    print
                print


def print_metadata(comm, tool=None):
    """Print metadata for tools used to annotate Communication
    """
    def _get_tokenizations(comm):
        tokenizations = []
        if comm.sectionList:
            for section in comm.sectionList:
                if section.sentenceList:
                    for sentence in section.sentenceList:
                        if sentence.tokenization:
                            tokenizations.append(sentence.tokenization)
        return tokenizations

    if tool is None or comm.metadata.tool == tool:
        print u"Communication:  %s\n" % comm.metadata.tool

    dependency_parse_tools = set()
    parse_tools = set()
    tokenization_tools = set()
    token_tagging_tools = set()
    for tokenization in _get_tokenizations(comm):
        tokenization_tools.add(tokenization.metadata.tool)
        if tokenization.tokenTaggingList:
            for tokenTagging in tokenization.tokenTaggingList:
                token_tagging_tools.add(tokenTagging.metadata.tool)
        if tokenization.dependencyParseList:
            for dependencyParse in tokenization.dependencyParseList:
                dependency_parse_tools.add(dependencyParse.metadata.tool)
        if tokenization.parseList:
            for parse in tokenization.parseList:
                parse_tools.add(parse.metadata.tool)

    communication_tagging_tools = set()
    for communication_tagging in lun(comm.communicationTaggingList):
        communication_tagging_tools.add(communication_tagging.metadata.tool)

    if tool is not None:
        dependency_parse_tools = dependency_parse_tools.intersection([tool])
        parse_tools = parse_tools.intersection([tool])
        tokenization_tools = tokenization_tools.intersection([tool])
        token_tagging_tools = token_tagging_tools.intersection([tool])
        communication_tagging_tools = communication_tagging_tools.intersection(
            [tool])

    if tokenization_tools:
        for toolname in sorted(tokenization_tools):
            print u"  Tokenization:  %s" % toolname
        print
    if dependency_parse_tools:
        for toolname in sorted(dependency_parse_tools):
            print u"    Dependency Parse:  %s" % toolname
        print
    if parse_tools:
        for toolname in sorted(parse_tools):
            print u"    Parse:  %s" % toolname
        print
    if token_tagging_tools:
        for toolname in sorted(token_tagging_tools):
            print u"    TokenTagging:  %s" % toolname
        print

    if comm.entityMentionSetList:
        for i, em_set in enumerate(comm.entityMentionSetList):
            if tool is None or em_set.metadata.tool == tool:
                print u"  EntityMentionSet #%d:  %s" % (
                    i, em_set.metadata.tool)
        print
    if comm.entitySetList:
        for i, entitySet in enumerate(comm.entitySetList):
            if tool is None or entitySet.metadata.tool == tool:
                print u"  EntitySet #%d:  %s" % (
                    i, entitySet.metadata.tool)
        print
    if comm.situationMentionSetList:
        for i, sm_set in enumerate(comm.situationMentionSetList):
            if tool is None or sm_set.metadata.tool == tool:
                print u"  SituationMentionSet #%d:  %s" % (
                    i, sm_set.metadata.tool)
        print
    if comm.situationSetList:
        for i, situationSet in enumerate(comm.situationSetList):
            if tool is None or situationSet.metadata.tool == tool:
                print u"  SituationSet #%d:  %s" % (
                    i, situationSet.metadata.tool)
        print

    if communication_tagging_tools:
        for toolname in sorted(communication_tagging_tools):
            print u"  CommunicationTagging:  %s" % toolname
        print


def print_sections(comm, tool=None):
    """Print information for all Sections, according to their spans.

    Args:

    - `comm`: A Concrete Communication
    """
    if tool is None or comm.metadata.tool == tool:
        text = comm.text
        for sect_idx, sect in enumerate(lun(comm.sectionList)):
            ts = sect.textSpan
            if ts is None:
                print u"Section %s does not have a textSpan "
                "field set" % (sect.uuid.uuidString)
                continue
            print u"Section %d (%s), from %d to %d:" % (
                sect_idx, sect.uuid.uuidString, ts.start, ts.ending)
            print u"%s" % (text[ts.start:ts.ending])
            print
        print


def print_situation_mentions(comm, tool=None):
    """Print information for all SituationMentions (some of which may
    not have Situations)

    Args:

    - `comm`: A Concrete Communication
    """
    for sm_set_idx, sm_set in enumerate(lun(comm.situationMentionSetList)):
        if tool is None or sm_set.metadata.tool == tool:
            print u"Situation Set %d (%s):" % (sm_set_idx,
                                               sm_set.metadata.tool)
            for sm_idx, sm in enumerate(sm_set.mentionList):
                print u"  SituationMention %d-%d:" % (sm_set_idx, sm_idx)
                _print_situation_mention(sm)
                print
            print


def print_situations(comm, tool=None):
    """Print information for all Situations and their SituationMentions

    Args:

    - `comm`: A Concrete Communication
    """
    for s_set_idx, s_set in enumerate(lun(comm.situationSetList)):
        if tool is None or s_set.metadata.tool == tool:
            print u"Situation Set %d (%s):" % (s_set_idx,
                                               s_set.metadata.tool)
            for s_idx, situation in enumerate(s_set.situationList):
                print u"  Situation %d-%d:" % (s_set_idx, s_idx)
                _p(6, 18, u"situationType", situation.situationType)
                for sm_idx, sm in enumerate(lun(situation.mentionList)):
                    print u" " * 6 + u"SituationMention %d-%d-%d:" % (
                        s_set_idx, s_idx, sm_idx)
                    _print_situation_mention(sm)
                print
            print


def _print_situation_mention(situationMention):
    """Helper function for printing info for a SituationMention"""
    if situationMention.text:
        _p(10, 20, u"text", situationMention.text)
    if situationMention.situationType:
        _p(10, 20, u"situationType", situationMention.situationType)
    for arg_idx, ma in enumerate(lun(situationMention.argumentList)):
        print u" " * 10 + u"Argument %d:" % arg_idx
        if ma.role:
            _p(14, 16, u"role", ma.role)
        if ma.entityMention:
            _p(14, 16, u"entityMention",
                u" ".join(_get_tokens_for_entityMention(ma.entityMention)))
        # A SituationMention can have an argumentList with a
        # MentionArgument that points to another SituationMention---
        # which could conceivably lead to loops.  We currently don't
        # traverse the list recursively, instead looking at only
        # SituationMentions referenced by top-level SituationMentions
        if ma.situationMention:
            print u" " * 14 + u"situationMention:"
            if situationMention.text:
                _p(18, 20, u"text", situationMention.text)
            if situationMention.situationType:
                _p(18, 20, u"situationType", situationMention.situationType)


def _p(indent_level, justified_width, fieldname, content):
    """Text alignment helper function"""
    print (
        (u" " * indent_level) +
        (fieldname + u":").ljust(justified_width) +
        content
    )


def print_text_for_communication(comm, tool=None):
    if tool is None or comm.metadata.tool == tool:
        print comm.text


def print_id_for_communication(comm, tool=None):
    if tool is None or comm.metadata.tool == tool:
        print comm.id


def print_communication_taggings_for_communication(comm, tool=None):
    communication_taggings = _filter_by_tool(
        lun(comm.communicationTaggingList), tool)
    for tagging in communication_taggings:
        print '%s: %s' % (
            tagging.taggingType,
            ' '.join('%s:%.3f' % p for p in
                     zip(tagging.tagList, tagging.confidenceList))
        )


def print_tokens_with_entityMentions(comm, tool=None):
    em_by_tkzn_id = _get_entityMentions_by_tokenizationId(
        comm, tool=tool)
    em_entity_num = _get_entity_number_for_entityMention_uuid(comm, tool=tool)
    tokenizations_by_section = _get_tokenizations_grouped_by_section(comm)

    for tokenizations_in_section in tokenizations_by_section:
        for tokenization in tokenizations_in_section:
            if tokenization.tokenList:
                text_tokens = [token.text
                               for token
                               in tokenization.tokenList.tokenList]
                u = tokenization.uuid.uuidString
                if u in em_by_tkzn_id:
                    for em in em_by_tkzn_id[u]:
                        first_token_index = em.tokens.tokenIndexList[0]
                        last_token_index = em.tokens.tokenIndexList[-1]
                        entity_number = em_entity_num[em.uuid.uuidString]
                        text_tokens[first_token_index] = (
                            u"<ENTITY ID=%d>%s" %
                            (entity_number, text_tokens[first_token_index])
                        )
                        text_tokens[last_token_index] = (
                            u"%s</ENTITY>" % text_tokens[last_token_index]
                        )
                print u" ".join(text_tokens)
        print


def print_tokens_for_communication(comm, tool=None):
    """
    """
    tokenizations_by_section = _get_tokenizations_grouped_by_section(
        comm, tool=tool)

    for tokenizations_in_section in tokenizations_by_section:
        for tokenization in tokenizations_in_section:
            if tokenization.tokenList:
                text_tokens = [token.text
                               for token
                               in tokenization.tokenList.tokenList]
                print u" ".join(text_tokens)
        print


def print_penn_treebank_for_communication(comm, tool=None):
    """Print Penn-Treebank parse trees for all tokenizations

    Args:

    - `comm`: A Concrete Communication object
    """
    tokenizations = _get_tokenizations(comm)

    for tokenization in tokenizations:
        if tokenization.parseList:
            for parse in tokenization.parseList:
                if tool is None or tool == parse.metadata.tool:
                    print penn_treebank_for_parse(parse) + u"\n\n"


def penn_treebank_for_parse(parse):
    """Get a Penn-Treebank style string for a Concrete Parse object

    Args:

    - `parse`: A Concrete Parse object

    Returns:

    - A string containing a Penn Treebank style parse tree representation
    """
    def _traverse_parse(nodes, node_index, indent=0):
        s = u""
        indent += len(nodes[node_index].tag) + 2
        if nodes[node_index].childList:
            s += u"(%s " % nodes[node_index].tag
            for i, child_node_index in enumerate(nodes[node_index].childList):
                if i > 0:
                    s += u"\n" + u" " * indent
                s += _traverse_parse(nodes, child_node_index, indent)
            s += u")"
        else:
            s += nodes[node_index].tag
        return s

    sorted_nodes = sorted(parse.constituentList, key=attrgetter('id'))
    return _traverse_parse(sorted_nodes, 0)


def _get_char_offset_tags_for_tokenization(comm, tokenization):
    if tokenization.tokenList:
        char_offset_tags = [None] * len(tokenization.tokenList.tokenList)

        if comm.text:
            for i, token in enumerate(tokenization.tokenList.tokenList):
                if token.textSpan:
                    char_offset_tags[i] = comm.text[
                        token.textSpan.start:token.textSpan.ending]
        return char_offset_tags


def _deps_for_tokenization(tokenization,
                           dependency_parse_index=0,
                           tool=None):
    '''
    Return a generator of the dependencies (Dependency objects) for
    a tokenization under the given tool.
    '''
    if tokenization.tokenList is not None:
        # Tokens that are not part of the dependency parse
        # (e.g. punctuation) are represented using an empty string
        dp_idx = _reconcile_index_and_tool(tokenization.dependencyParseList,
                                           dependency_parse_index,
                                           tool)

        if _valid_index_lun(tokenization.dependencyParseList, dp_idx):
            dp = tokenization.dependencyParseList[dp_idx]
            if tool is None or dp.metadata.tool == tool:
                for dependency in dp.dependencyList:
                    yield dependency


def _sorted_dep_list_for_tokenization(tokenization,
                                      dependency_parse_index=0,
                                      tool=None):
    '''
    Return output of _deps_for_tokenization in a list whose length
    is equal to the number of tokens in this tokenization's token list,
    where the element at index i is a dependency if there is a
    dependency whose dep field is i and None otherwise.
    '''
    if tokenization.tokenList is not None:
        dep_list = [None] * len(tokenization.tokenList.tokenList)
        for dep in _deps_for_tokenization(
                tokenization, dependency_parse_index=dependency_parse_index,
                tool=tool):
            dep_list[dep.dep] = dep
        return dep_list
    else:
        return []


def _get_conll_head_tags_for_tokenization(tokenization,
                                          dependency_parse_index=0,
                                          tool=None):
    """Get a list of ConLL 'HEAD tags' for a tokenization

    In the ConLL data format:

        http://ufal.mff.cuni.cz/conll2009-st/task-description.html

    the HEAD for a token is the (1-indexed) index of that token's
    parent token.  The root token of the dependency parse has a HEAD
    index of 0.

    Args:

    - `tokenization`: A Concrete Tokenization object

    Returns:

    - A list of ConLL 'HEAD tag' strings, with one HEAD tag for each
      token in the supplied tokenization.  If a token does not have
      a HEAD tag (e.g. punctuation tokens), the HEAD tag is an empty
      string.

      If the tokenization does not have a Dependency Parse, this
      function returns a list of empty strings for each token in the
      supplied tokenization.
    """
    return map(
        lambda dep: '' if dep is None else (
            0 if dep.gov is None else dep.gov + 1),
        _sorted_dep_list_for_tokenization(
            tokenization, dependency_parse_index=dependency_parse_index,
            tool=tool))


def _get_conll_deprel_tags_for_tokenization(tokenization,
                                            dependency_parse_index=0,
                                            tool=None):
    """Get a list of ConLL 'DEPREL tags' for a tokenization

    In the ConLL data format:

        http://ufal.mff.cuni.cz/conll2009-st/task-description.html

    the DEPREL for a token is the type of that token's dependency with
    its parent.

    Args:

    - `tokenization`: A Concrete Tokenization object

    Returns:

    - A list of ConLL 'DEPREL tag' strings, with one DEPREL tag for each
      token in the supplied tokenization.  If a token does not have
      a DEPREL tag (e.g. punctuation tokens), the DEPREL tag is an empty
      string.

      If the tokenization does not have a Dependency Parse, this
      function returns a list of empty strings for each token in the
      supplied tokenization.
    """
    return map(
        lambda dep: '' if dep is None else (
            '' if dep.edgeType is None else dep.edgeType),
        _sorted_dep_list_for_tokenization(
            tokenization, dependency_parse_index=dependency_parse_index,
            tool=tool))


def _get_entityMentions_by_tokenizationId(comm, tool=None):
    """Get entity mentions for a Communication grouped by Tokenization
    UUID string

    Args:

    - `comm`: A Concrete Communication object

    Returns:

    - A dictionary of lists of EntityMentions, where the dictionary
      keys are Tokenization UUID strings.
    """
    mentions_by_tkzn_id = defaultdict(list)
    for entitySet in lun(comm.entitySetList):
        for entity in entitySet.entityList:
            for entityMention in entity.mentionList:
                if (tool is None or
                        entityMention.entityMentionSet.metadata.tool == tool):
                    u = entityMention.tokens.tokenizationId.uuidString
                    mentions_by_tkzn_id[u].append(entityMention)
    return mentions_by_tkzn_id


def _get_entity_number_for_entityMention_uuid(comm, tool=None):
    """Create mapping from EntityMention UUID to (zero-indexed)
    'Entity Number'

    Args:

    - `comm`: A Concrete Communication object

    Returns:

    - A dictionary where the keys are EntityMention UUID strings,
      and the values are "Entity Numbers", where the first Entity is
      assigned number 0, the second Entity is assigned number 1,
      etc.
    """
    entity_number_for_entityMention_uuid = {}
    entity_number_counter = 0

    if comm.entitySetList:
        for entitySet in comm.entitySetList:
            for entity in entitySet.entityList:
                any_mention = False
                for entityMention in entity.mentionList:
                    if (tool is None or
                            entityMention.entityMentionSet.metadata.tool ==
                            tool):
                        entity_number_for_entityMention_uuid[
                            entityMention.uuid.uuidString
                        ] = entity_number_counter
                    any_mention = True
                if any_mention:
                    entity_number_counter += 1
    return entity_number_for_entityMention_uuid


def _get_lemma_tags_for_tokenization(tokenization, lemma_tokentagging_index=0,
                                     tool=None):
    """Get lemma tags for a tokenization

    Args:

    - `tokenization`: A Concrete Tokenization object

    Returns:

    - A list of lemma tags for each token in the Tokenization
    """
    if tokenization.tokenList:
        lemma_tags = [""] * len(tokenization.tokenList.tokenList)
        lemma_tts = _get_tokentaggings_of_type(tokenization, u"lemma",
                                               tool=tool)
        if (lemma_tts and
                len(lemma_tts) > lemma_tokentagging_index):
            tag_for_tokenIndex = {}
            tt = lemma_tts[lemma_tokentagging_index]
            for taggedToken in tt.taggedTokenList:
                tag_for_tokenIndex[taggedToken.tokenIndex] = taggedToken.tag
            for i, token in enumerate(tokenization.tokenList.tokenList):
                if i in tag_for_tokenIndex:
                    lemma_tags[i] = tag_for_tokenIndex[i]
        return lemma_tags


def _get_ner_tags_for_tokenization(tokenization, ner_tokentagging_index=0,
                                   tool=None):
    """Get Named Entity Recognition tags for a tokenization

    Args:

    - `tokenization`: A Concrete Tokenization object

    Returns:

    - A list of NER tags for each token in the Tokenization
    """
    if tokenization.tokenList:
        ner_tags = [""] * len(tokenization.tokenList.tokenList)
        ner_tts = _get_tokentaggings_of_type(tokenization, u"NER", tool=tool)
        if (ner_tts and
                len(ner_tts) > ner_tokentagging_index):
            tag_for_tokenIndex = {}
            for taggedToken in ner_tts[ner_tokentagging_index].taggedTokenList:
                tag_for_tokenIndex[taggedToken.tokenIndex] = taggedToken.tag
            for i, token in enumerate(tokenization.tokenList.tokenList):
                try:
                    ner_tags[i] = tag_for_tokenIndex[i]
                except IndexError:
                    ner_tags[i] = u""
                if ner_tags[i] == u"NONE":
                    ner_tags[i] = u""
        return ner_tags


def _get_pos_tags_for_tokenization(tokenization, pos_tokentagging_index=0,
                                   tool=None):
    """Get Part-of-Speech tags for a tokenization

    Args:

    - tokenization: A Concrete Tokenization object

    Returns:

    - A list of POS tags for each token in the Tokenization
    """
    if tokenization.tokenList:
        pos_tags = [""] * len(tokenization.tokenList.tokenList)
        pos_tts = _get_tokentaggings_of_type(tokenization, u"POS", tool=tool)
        if pos_tts and len(pos_tts) > pos_tokentagging_index:
            tag_for_tokenIndex = {}
            for taggedToken in pos_tts[pos_tokentagging_index].taggedTokenList:
                tag_for_tokenIndex[taggedToken.tokenIndex] = taggedToken.tag
            for i, token in enumerate(tokenization.tokenList.tokenList):
                if i in tag_for_tokenIndex:
                    pos_tags[i] = tag_for_tokenIndex[i]
        return pos_tags


def _get_tokenizations(comm, tool=None):
    """Returns a flat list of all Tokenization objects in a Communication

    Args:

    - `comm`: A Concrete Communication

    Returns:

    - A list of all Tokenization objects within the Communication
    """
    tokenizations = []

    if comm.sectionList:
        for section in comm.sectionList:
            if section.sentenceList:
                for sentence in section.sentenceList:
                    if sentence.tokenization:
                        if (tool is None or
                                sentence.tokenization.metadata.tool == tool):
                            tokenizations.append(sentence.tokenization)
    return tokenizations


def _get_tokenizations_grouped_by_section(comm, tool=None):
    """Returns a list of lists of Tokenization objects in a Communication

    Args:

    - `comm`: A Concrete Communication

    Returns:

    - Returns a list of lists of Tokenization objects, where the
      Tokenization objects are grouped by Section
    """
    tokenizations_by_section = []

    if comm.sectionList:
        for section in comm.sectionList:
            tokenizations_in_section = []
            if section.sentenceList:
                for sentence in section.sentenceList:
                    if sentence.tokenization:
                        if (tool is None or
                                sentence.tokenization.metadata.tool == tool):
                            tokenizations_in_section.append(
                                sentence.tokenization)
            tokenizations_by_section.append(tokenizations_in_section)

    return tokenizations_by_section


def _get_tokens_for_entityMention(entityMention):
    """Get list of token strings for an EntityMention

    Args:

    - `entityMention`: A Concrete EntityMention argument

    Returns:

    - A list of token strings
    """
    tokens = []
    for tokenIndex in entityMention.tokens.tokenIndexList:
        tokens.append(entityMention.tokens.tokenization.tokenList.tokenList[
                      tokenIndex].text)
    return tokens


def _get_tokentaggings_of_type(tokenization, taggingType, tool=None):
    """Returns a list of TokenTagging objects with the specified taggingType

    Args:

    - `tokenization`: A Concrete Tokenizaiton object
    - `taggingType`: A string value for the specified TokenTagging.taggingType

    Returns:

    - A list of TokenTagging objects
    """
    return [
        tt for tt in tokenization.tokenTaggingList
        if tt.taggingType.lower() == taggingType.lower() and (
            tool is None or tt.metadata.tool == tool)
    ]



from concrete.inspect import (
    _reconcile_index_and_tool,
    _valid_index_lun,
    _get_conll_deprel_tags_for_tokenization as cdt
)

from concrete import (AnnotationMetadata,
                      Dependency,
                      DependencyParse,
                      Tokenization,
                      TokenizationKind,
                      Token,
                      TokenList,
                      UUID)

import time


class ReconcileIndexAndTool:

    default_tool = "My awesome tool"

    def __init__(self, num, t=default_tool):
        self.lst = [
            DependencyParse(
                uuid=UUID(uuidString=unicode(1)),
                metadata=AnnotationMetadata(
                    timestamp=int(time.time()),
                    tool=t if i == 0 else u"tool_" + unicode(i)
                ),
                dependencyList=[]
            ) for i in xrange(num)
        ] if num is not None else None


def test_reconcile_index_and_tool_none_list():
    lst = ReconcileIndexAndTool(None).lst
    assert _reconcile_index_and_tool(lst, 0, None) == -1
    assert _reconcile_index_and_tool(lst, -2, None) == -1
    assert _reconcile_index_and_tool(lst, 0, "My awesome tool") == -1


def test_reconcile_index_and_tool_empty_list():
    lst = ReconcileIndexAndTool(0).lst
    assert _reconcile_index_and_tool(lst, 0, None) == -1
    assert _reconcile_index_and_tool(lst, -2, None) == -1
    assert _reconcile_index_and_tool(lst, 0, "My awesome tool") == -1


def test_reconcile_index_and_tool_singleton_list():
    lst = ReconcileIndexAndTool(1).lst
    assert _reconcile_index_and_tool(lst, 0, None) == 0
    assert _reconcile_index_and_tool(lst, -2, None) == -2
    assert _reconcile_index_and_tool(lst, 0, "My awesome tool") == 0
    assert _reconcile_index_and_tool(lst, 0, "Your awesome tool") == -1


def test_reconcile_index_and_tool_three_list():
    lst = ReconcileIndexAndTool(3).lst
    assert _reconcile_index_and_tool(lst, 0, None) == 0
    assert _reconcile_index_and_tool(lst, -2, None) == -2
    assert _reconcile_index_and_tool(lst, 0, "My awesome tool") == 0
    assert _reconcile_index_and_tool(lst, 0, "Your awesome tool") == -1
    tmp = lst[0]
    lst[0] = lst[2]
    lst[2] = tmp
    assert _reconcile_index_and_tool(lst, 0, "My awesome tool") == 2
    assert _reconcile_index_and_tool(lst, 3, "My awesome tool") == 2
    assert _reconcile_index_and_tool(lst, 3, "Your awesome tool") == -1


def test_valid_index_lun_none_list():
    assert not _valid_index_lun(None, 0)
    assert not _valid_index_lun(None, -2)
    assert not _valid_index_lun(None, 0)


def test_valid_index_lun_empty_list():
    assert not _valid_index_lun([], 0)
    assert not _valid_index_lun([], -2)
    assert not _valid_index_lun([], 0)


def test_valid_index_lun_singleton_list():
    assert _valid_index_lun(xrange(1), 0)
    assert not _valid_index_lun(xrange(1), -2)
    assert not _valid_index_lun(xrange(1), 1)


class FakeTokenizationAndDependency:

    default_tool = "My awesome tool"

    def __init__(self, num, num_dp=1, t=default_tool):
        self.tokenization = Tokenization(
            uuid=UUID(uuidString=u"id"),
            metadata=AnnotationMetadata(
                timestamp=int(time.time()),
                tool=u"tool"
            ),
            kind=TokenizationKind.TOKEN_LIST,
        )
        if num is not None:
            self.tokenization.tokenList = TokenList(
                tokenList=[Token(tokenIndex=i,
                                 text=unicode(i))
                           for i in xrange(num)]
            )
            self.tokenization.dependencyParseList = [
                DependencyParse(
                    uuid=UUID(uuidString=u"depparse_" + unicode(num)),
                    metadata=AnnotationMetadata(
                        timestamp=int(time.time()),
                        tool=t if dp_idx == 0 else
                        (u"dptool" + unicode(dp_idx))
                    ),
                    dependencyList=[
                        Dependency(
                            gov=i-1, dep=i,
                            edgeType=u"edge_" + unicode(i) +
                            "/" + unicode(dp_idx)
                        ) for i in xrange(num)
                    ]
                ) for dp_idx in xrange(num_dp)
            ]


def test_get_conll_deprel_tags_none_tokens():
    tokenization = FakeTokenizationAndDependency(None).tokenization
    for i in xrange(-1, 2):
        for tool in (None, "", FakeTokenizationAndDependency.default_tool):
            assert cdt(tokenization, i, tool) == []


def test_get_conll_deprel_tags_zero_tokens():
    tokenization = FakeTokenizationAndDependency(0).tokenization
    for i in xrange(-1, 2):
        for tool in (None, "", FakeTokenizationAndDependency.default_tool):
            assert cdt(tokenization, i, tool) == []


def test_get_conll_deprel_tags_one_tokens_found():
    tokenization = FakeTokenizationAndDependency(1).tokenization
    assert cdt(tokenization, 0, None) == [u"edge_0/0"]
    assert (
        cdt(tokenization, 0, FakeTokenizationAndDependency.default_tool) ==
        [u"edge_0/0"])
    assert cdt(tokenization, 0, "") == [""]

    for tool in (None, ""):
        assert cdt(tokenization, -1, tool) == [""]

    assert (
        cdt(tokenization, -1, FakeTokenizationAndDependency.default_tool) ==
        [u"edge_0/0"])

    assert (
        cdt(tokenization, 1, FakeTokenizationAndDependency.default_tool) ==
        [u"edge_0/0"])
    for tool in (None, ""):
        assert cdt(tokenization, 1, tool) == [""]


def test_get_conll_deprel_tags_arbitrary_num_tokens_not_found():
    t = "Other great tool"
    for num_tokens in xrange(1, 10):
        tokenization = FakeTokenizationAndDependency(num_tokens,
                                                     t=t).tokenization
        assert cdt(tokenization, 0, None) == [u"edge_" + unicode(i) + "/0"
                                              for i in xrange(num_tokens)]
        assert (
            cdt(tokenization, 0, FakeTokenizationAndDependency.default_tool) ==
            [""] * num_tokens
        )
        assert cdt(tokenization, 0, "") == [""] * num_tokens

        assert cdt(tokenization, -1, None) == [""] * num_tokens
        assert tokenization.tokenList

        assert cdt(tokenization, 0, None) == [u"edge_" + unicode(j) + "/0"
                                              for j in xrange(num_tokens)]
        for tool in ("", FakeTokenizationAndDependency.default_tool):
            assert cdt(tokenization, 0, tool) == [""] * num_tokens

        for i in xrange(1, num_tokens+1):
            for tool in (None, "",
                         FakeTokenizationAndDependency.default_tool):
                assert cdt(tokenization, i, tool) == [""] * num_tokens


def test_get_conll_deprel_tags_arbitrary_num_tokens_num_dp_not_found():
    t = "Other great tool"
    # without this memoization, style fails
    style_tool = FakeTokenizationAndDependency.default_tool
    for num_dp in xrange(1, 4):
        for num_tokens in xrange(1, 10):
            tokenization = FakeTokenizationAndDependency(num_tokens,
                                                         num_dp=num_dp,
                                                         t=t).tokenization

            # no tool given, with valid index: return indexed parse
            assert cdt(tokenization, 0, None) == [u"edge_" + unicode(i) + "/0"
                                                  for i in xrange(num_tokens)]
            # tools given, with valid index, but not found
            assert cdt(tokenization, 0, style_tool) == [""] * num_tokens
            assert cdt(tokenization, 0, "") == [""] * num_tokens

            # invalid index, no tool given: return empty
            assert cdt(tokenization, -1, None) == [""] * num_tokens

            for i in xrange(num_dp):
                # no tool given, valid index: return specified parse
                assert cdt(tokenization, i, None) == [
                    u"edge_" + unicode(j) + "/" + unicode(i)
                    for j in xrange(num_tokens)]
                # tool given, valid index, not found: return empty
                for tool in ("",
                             FakeTokenizationAndDependency.default_tool):
                    assert cdt(tokenization, i, tool) == [""] * num_tokens

            # invalid index: return empty
            for tool in (None, "",
                         FakeTokenizationAndDependency.default_tool):
                assert cdt(tokenization, num_dp, tool) == [""] * num_tokens


def test_get_conll_deprel_tags_arbitrary_num_tokens_num_dp_found():
    # sigh... memoize to let style checks pass
    style_tool = FakeTokenizationAndDependency.default_tool
    for num_dp in xrange(1, 4):
        for num_tokens in xrange(1, 10):
            tokenization = FakeTokenizationAndDependency(
                num_tokens,
                num_dp=num_dp).tokenization

            # no tool given, with valid index: return indexed parse
            assert cdt(tokenization, 0, None) == [u"edge_" + unicode(i) + "/0"
                                                  for i in xrange(num_tokens)]
            # tool given, with valid index, not found
            assert cdt(tokenization, 0, style_tool) == [
                u"edge_" + unicode(i) + "/0" for i in xrange(num_tokens)]
            assert cdt(tokenization, 0, "") == [""] * num_tokens

            # invalid index, no tool given: return empty
            assert cdt(tokenization, -1, None) == [""] * num_tokens

            for i in xrange(num_dp):
                # tool given, valid index, not found: return empty
                assert cdt(tokenization, i, "") == [""] * num_tokens
                # tool none, valid index, found: return requested
                assert cdt(tokenization, i, None) == [
                    u"edge_" + unicode(j) + "/" + unicode(i)
                    for j in xrange(num_tokens)]
                # tool given, valid index, found: return correct
                assert cdt(tokenization, i, style_tool) == [
                    u"edge_" + unicode(j) + "/0" for j in xrange(num_tokens)]

            # invalid index: return empty
            for tool in (None, ""):
                assert cdt(tokenization, num_dp, tool) == [""] * num_tokens

            assert cdt(tokenization, num_dp, style_tool) == [
                u"edge_" + unicode(j) + "/0" for j in xrange(num_tokens)]



import unittest

from concrete.util.comm_container import (
    DirectoryBackedCommunicationContainer,
    MemoryBackedCommunicationContainer,
    ZipFileBackedCommunicationContainer,
)

from concrete.validate import validate_communication


class TestDirectoryBackedCommunicationContainer(unittest.TestCase):
    def test_find_files_recursively(self):
        directory_path = u'tests/testdata/a'
        cc = DirectoryBackedCommunicationContainer(directory_path)
        self.assertEqual(3, len(cc))

    def test_retrieve(self):
        directory_path = u'tests/testdata/a'
        cc = DirectoryBackedCommunicationContainer(directory_path)
        self.assertEqual(3, len(cc))
        self.assertTrue(u'simple_1' in cc)
        for comm_id in cc:
            comm = cc[comm_id]
            self.assertTrue(validate_communication(comm))


class TestMemoryBackedCommunicationContainer(unittest.TestCase):
    def test_filesize_check(self):
        comm_path = u'tests/testdata/simple.tar.gz'
        with self.assertRaises(Exception):
            MemoryBackedCommunicationContainer(comm_path, max_file_size=500)

    def test_retrieve(self):
        comm_path = u'tests/testdata/simple.tar.gz'
        cc = MemoryBackedCommunicationContainer(comm_path)
        self.assertEqual(3, len(cc))
        self.assertTrue(u'one' in cc)
        for comm_id in cc:
            comm = cc[comm_id]
            self.assertTrue(validate_communication(comm))


class TestZipFileBackedCommunicationContainer(unittest.TestCase):
    def test_retrieve(self):
        zipfile_path = u'tests/testdata/simple.zip'
        cc = ZipFileBackedCommunicationContainer(zipfile_path)
        self.assertEqual(3, len(cc))
        self.assertTrue(u'simple_1' in cc)
        for comm_id in cc:
            comm = cc[comm_id]
            self.assertTrue(validate_communication(comm))



from setuptools import setup
import glob

execfile('concrete/version.py')

if __name__ == '__main__':
    setup(
        name="concrete",
        version=__version__,
        description="Python modules and scripts for working with Concrete",

        packages=[
            'concrete',

            # Python code generated by Thrift Compiler
            'concrete.access',
            'concrete.annotate',
            'concrete.audio',
            'concrete.clustering',
            'concrete.communication',
            'concrete.email',
            'concrete.entities',
            'concrete.exceptions',
            'concrete.language',
            'concrete.learn',
            'concrete.linking',
            'concrete.metadata',
            'concrete.nitf',
            'concrete.search',
            'concrete.services',
            'concrete.services.results',
            'concrete.situations',
            'concrete.spans',
            'concrete.structure',
            'concrete.twitter',
            'concrete.uuid',

            # Python code generated by people
            'concrete.util',
        ],

        scripts=glob.glob('scripts/*')
                  + glob.glob('concrete/services/*-remote'),

        setup_requires=['pytest-runner'],
        tests_require=['pytest', 'pytest-mock', 'pytest-cov', 'flake8'],

        install_requires=[
            'humanfriendly',
            'networkx',
            'testfixtures',
            'thrift==0.10.0',
            'redis>=2.10.0',
            'pycountry==1.20',
            'requests',
        ],

        url="https://github.com/hltcoe/concrete-python",
        license="BSD",
    )
